import polars as pl
import numpy as np

from ._voxelize import (
        Sphere, Atom, Grid,
        _add_atom_to_image, _get_voxel_center_coords,
)
from dataclasses import dataclass

from typing import TypeAlias
from numpy.typing import NDArray

"""\
Data structures and naming conventions
======================================
This list only includes data types that don't have their own classes.

`img`:
  A `np.ndarray` that contains the voxelized scene.

`atoms`:
  A `pandas.DataFrame` with the following columns: x, y, z, element.  This is 
  the way atoms are expected to be represented externally.  Internally, the 
  Atom data class is used to represent individual atoms.

`voxel`:
  A 3D `numpy.ndarray` containing the index for one of the cells in the image.  
  Generally, these indices are constrained to actually fall within the image in 
  question (i.e. no indices greater than the size of the image or less than 0).  
  Note that a `grid` object is needed to determine the physical location of the 
  voxel.  When multiple voxels are involved, the array has dimensions of (N,3).

`coords`:
  A 3D `numpy.ndarray` containing the location of the center of a voxel in 
  real-world coordinates, in units of Angstroms.  When multiple coordinates 
  that involved, the array has dimensions of (N,3).
"""

@dataclass
class ImageParams:
    channels: int
    grid: Grid

Image: TypeAlias = NDArray

def image_from_atoms(atoms: pl.DataFrame, img_params: ImageParams) -> Image:
    _check_channels(atoms, img_params.channels)

    img = _make_empty_image(img_params)
    grid = img_params.grid

    # Without this filter, `_find_voxels_possibly_contacting_sphere()` becomes 
    # a performance bottleneck.
    atoms = _discard_atoms_outside_image(atoms, grid)

    for row in atoms.iter_rows(named=True):
        atom = _make_atom(row)
        _add_atom_to_image(img, grid, atom)

    return img
        
def set_atom_radius_A(atoms: pl.DataFrame, radius_A: float):
    """\
    Assign all atoms the same radius.

    Arguments:
        atoms:
            A dataframe representing the atoms to voxelize.

        radius_A:
            The radius to assign, in angstroms.

    Returns:
        The input dataframe, with a new *radius_A* column.  Every row in this 
        column will have the same value.
    """
    return atoms.with_columns(radius_A=radius_A)

def set_atom_channels_by_element(
        atoms: pl.DataFrame,
        channels: list[str],
        first_match: bool = False,
        allow_missing_atoms: bool = False,
):
    """\
    Assign atoms to channels based on their element types.

    Arguments:
        atoms:
            A dataframe representing the atoms to voxelize.  This function 
            requires a column named "element", which must contain element names 
            as strings.

        channels:
            A list of regular expression patterns, each of which should match 
            elements that belong in a particular channel.  The *first_match* 
            argument controls what happens if multiple patterns match the same 
            atom.

            For example, ``['C', 'N', 'O', '.*']`` would place carbon in the 
            first channel, nitrogen in the second channel, and oxygen in the 
            third channel.  The fourth channel would contain all atoms 
            (including those already in one of the first three channels) if 
            *first_match=False*, otherwise it contain just those atoms that 
            aren't in of the earlier channels.

        first_match:
            If *True*, only allow each atom to occupy one channel: the first 
            one it matches.  If *False*, each atom will occupy every channel it 
            matches.

        allow_missing_atoms:
            If *True*, atoms that aren't assigned to any channel will be 
            silently removed.  By default, an error will be raised if any such 
            atoms are present.

    Returns:
        The input dataframe, with a *channels* column added.  Each entry in 
        this column will be a list of integers, where each integer identifies a 
        single channel and will be in the range [0, ``len(channels) - 1``].
    """
    channel_exprs = [
            pl.when(
                pl.col('element').str.contains(f'^({pattern})$')
            )
            .then(i)
            for i, pattern in enumerate(channels)
    ]
    if first_match:
        channel_exprs = pl.coalesce(channel_exprs)

    atoms = (
            atoms
            .with_columns(
                channels=pl.concat_list(channel_exprs).list.drop_nulls()
            )
    )

    if allow_missing_atoms:
        return atoms.filter(pl.col('channels').list.len() > 0)

    have_missing_atoms = (
            atoms
            .select(
                (pl.col('channels').list.len() == 0).any()
            )
            .item()
    )
    if have_missing_atoms:
        missing_elements = (
                atoms
                .filter(pl.col('channels').list.len() == 0)
                .get_column('element')
                .unique()
                .to_list()
        )
        raise ValidationError(f"""\
all atoms must be assigned at least one channel
• channels: {channels!r}
✖ unassigned elements: {missing_elements!r}
""")

    return atoms

def get_voxel_center_coords(grid, voxels):
    # There are two things to keep in mind when passing arrays between 
    # python/numpy and C++/Eigen:
    #
    # - Coordinates and voxel indices are represented as row vectors by 
    #   python/numpy, and as column vectors by C++/Eigen.  This means that 
    #   arrays have to be transposed when moving from one language to the 
    #   other.  In principle, it would be possible to use the same row/column 
    #   vector convention in both languages.  But this would make it harder to 
    #   interact with third-party libraries like `overlap`.
    #
    # - Eigen doesn't have 1D arrays.  Instead it has vectors, which are just 
    #   2D matrices with either 1 row or 1 column.  When converting a vector 
    #   from C++/Eigen back to python/numpy, it's not clear whether the 
    #   resulting array should be 1D or 2D.  This ambiguity can be resolved by 
    #   looking at the shape of the original numpy input.
    #
    # I decided against accounting for either of these issues in the binding 
    # code itself.  The main reason for exposing most of the C++ functions to 
    # python is testing, and for that it's not helpful to be changing the 
    # inputs and outputs.  But this specific function is useful in other 
    # contexts, so I wrote this wrapper function to enforce the python 
    # conventions.

    coords_A = _get_voxel_center_coords(grid, voxels.T).T
    return coords_A.reshape(voxels.shape)


def _check_channels(atoms, num_channels):
    channels = atoms['channels'].explode()
    assert channels.min() >= 0
    assert channels.max() < num_channels

def _make_empty_image(img_params):
    shape = img_params.channels, *img_params.grid.shape
    return np.zeros(shape, dtype=np.float32)

def _discard_atoms_outside_image(atoms, grid):
    max_r = atoms['radius_A'].max()

    min_corner = grid.center_A - (grid.length_A / 2 + max_r)
    max_corner = grid.center_A + (grid.length_A / 2 + max_r)

    return atoms.filter(
            pl.col('x') > min_corner[0],
            pl.col('x') < max_corner[0],
            pl.col('y') > min_corner[1],
            pl.col('y') < max_corner[1],
            pl.col('z') > min_corner[2],
            pl.col('z') < max_corner[2],
    )

def _make_atom(row):
    return Atom(
            sphere=Sphere(
                center_A=np.array([row['x'], row['y'], row['z']]),
                radius_A=row['radius_A'],
            ),
            channels=row['channels'],
            occupancy=row['occupancy'],
    )

class ValidationError(Exception):
    pass



from skbuild import setup, exceptions, cmaker
from packaging.version import Version

# Add CMake as a build requirement if no or only an outdated version is
# available.
setup_requires = []
try:
    if Version(cmaker.get_cmake_version()) < Version('3.12'):
        setup_requires.append('cmake')
except exceptions.SKBuildError:
    setup_requires.append('cmake')

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

VERSION = '0.1.1dev0'

setup(
    name='overlap',
    version=VERSION,
    author='Severin Strobl',
    author_email='git@severin-strobl.de',
    description='Exact calculation of the overlap volume and area of spheres '
                'and mesh elements',
    url='https://github.com/severinstrobl/overlap',
    license='GPLv3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['overlap'],
    package_dir={'overlap': 'python'},
    python_requires='>=3.6',
    setup_requires=setup_requires,
    install_requires=[
        'numpy'
    ],
    cmake_install_dir='python',
    cmake_args=[f"-DSKBUILD_PROJECT_VERSION={VERSION}"],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Topic :: Scientific/Engineering'
    ]
)



# Exact calculation of the overlap volume of spheres and mesh elements.
# http://dx.doi.org/10.1016/j.jcp.2016.02.003

# Copyright (C) 2021 Severin Strobl <git@severin-strobl.de>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import overlap


def test_sphere():
    s = overlap.Sphere((1, 1, 1), 1)
    np.testing.assert_almost_equal(s.center, [1, 1, 1])
    np.testing.assert_almost_equal(s.radius, 1.0)
    np.testing.assert_almost_equal(s.volume, 4./3. * np.pi)


class TestElements:
    def test_tetrahedron(self):
        vertices = np.array((
            (-1, -np.sqrt(1./3.), -np.sqrt(1./6.)),
            (1, -np.sqrt(1./3.), -np.sqrt(1./6.)),
            (0, np.sqrt(4./3.), -np.sqrt(1./6.)),
            (0, 0, np.sqrt(3./2.))
        ))
        tet = overlap.Tetrahedron(vertices)

        np.testing.assert_almost_equal(tet.volume, 8. / (6 * np.sqrt(2)))
        np.testing.assert_almost_equal(tet.center, [0, 0, 0])
        np.testing.assert_almost_equal(tet.surface_area, 4 * np.sqrt(3))

    def test_wedge(self):
        vertices = [
            (-1, -1, -1), (1, -1, -1), (1,  1, -1),
            (-1, -1,  1), (1, -1,  1), (1,  1,  1),
        ]

        wedge = overlap.Wedge(vertices)

        np.testing.assert_almost_equal(wedge.volume, 4.0)
        np.testing.assert_almost_equal(wedge.center, [1./3., -1./3., 0])
        np.testing.assert_almost_equal(wedge.surface_area,
                                       12.0 + 2 * np.sqrt(8))

    def test_hexahedron(self):
        vertices = [
            (-1, -1, -1), (1, -1, -1), (1,  1, -1), (-1,  1, -1),
            (-1, -1,  1), (1, -1,  1), (1,  1,  1), (-1,  1,  1)
        ]

        hexa = overlap.Hexahedron(vertices)

        np.testing.assert_almost_equal(hexa.volume, 8.0)
        np.testing.assert_almost_equal(hexa.center, [0, 0, 0])
        np.testing.assert_almost_equal(hexa.surface_area, 24.0)


class TestOverlap:
    def test_wedge(self):
        vertices = [
            (-1, -1, -1), (1, -1, -1), (1,  1, -1),
            (-1, -1,  1), (1, -1,  1), (1,  1,  1),
        ]

        wedge = overlap.Wedge(vertices)
        s = overlap.Sphere((1, 1, 1), 1)

        np.testing.assert_almost_equal(overlap.overlap(s, wedge), np.pi / 12)

    def test_hexahedron(self):
        vertices = [
            (-1, -1, -1), (1, -1, -1), (1,  1, -1), (-1,  1, -1),
            (-1, -1,  1), (1, -1,  1), (1,  1,  1), (-1,  1,  1)
        ]

        hexa = overlap.Hexahedron(vertices)
        s = overlap.Sphere((1, 1, 1), 1)

        np.testing.assert_almost_equal(overlap.overlap(s, hexa), np.pi / 6)


class TestOverlapArea:
    def test_wedge(self):
        vertices = [
            (-1, -1, -1), (1, -1, -1), (1,  1, -1),
            (-1, -1,  1), (1, -1,  1), (1,  1,  1),
        ]

        wedge = overlap.Wedge(vertices)
        s = overlap.Sphere((0, 0, 0), 1)

        np.testing.assert_almost_equal(overlap.overlap_area(s, wedge), [
            s.surface_area / 2,
            0., 0., 0., s.radius**2 * np.pi, 0.,
            s.radius**2 * np.pi
        ])

    def test_hexahedron(self):
        s = overlap.Sphere((1, 1, 1), 1)

        vertices = [
            (-1, -1, -1), (1, -1, -1), (1,  1, -1), (-1,  1, -1),
            (-1, -1,  1), (1, -1,  1), (1,  1,  1), (-1,  1,  1)
        ]

        hexa = overlap.Hexahedron(vertices)

        face_overlap = s.radius**2 * np.pi / 4

        np.testing.assert_almost_equal(overlap.overlap_area(s, hexa), [
            s.surface_area / 8,
            0., 0., face_overlap, face_overlap, 0, face_overlap,
            3 * face_overlap
        ])



import pymol
import polars as pl
import numpy as np
import macromol_dataframe as mmdf
import mixbox

from pymol import cmd
from pymol.cgo import (
        ALPHA, BEGIN, COLOR, CONE, CYLINDER, END, LINES, NORMAL, TRIANGLE_FAN,
        VERTEX,
)
from itertools import product
from pipeline_func import f

from .voxelize import (
        ImageParams, Grid, image_from_atoms, set_atom_radius_A,
        set_atom_channels_by_element, get_voxel_center_coords,
)

def voxelize(
        center_sele=None,
        all_sele='all',
        length_voxels=21,
        resolution_A=0.75,
        channels='C,N,O',
        element_radius_A=None,
        outline=False,
        state=-1,
        sele_name='within_img',
        obj_name='voxels',
        outline_name='outline',
):
    atoms = mmdf.from_pymol(all_sele, state)
    length_voxels = int(length_voxels)
    resolution_A = float(resolution_A)
    center_A = np.array(cmd.centerofmass(center_sele or all_sele, state))
    channels = parse_channels(channels)
    element_radius_A = parse_element_radius_A(element_radius_A, resolution_A)
    state = int(state)

    atoms = (
            atoms
            | f(set_atom_channels_by_element, channels, first_match=True)
            | f(set_atom_radius_A, element_radius_A)
    )
    img_params = ImageParams(
            grid=Grid(
                length_voxels=length_voxels,
                resolution_A=resolution_A,
                center_A=center_A,
            ),
            channels=len(channels),
    )
    select_view(
            sele_name,
            all_sele,
            img_params.grid,
    )
    render_view(
            obj_names=dict(
                voxels=obj_name,
                outline=outline_name,
            ),
            atoms_i=atoms,
            img_params=img_params,
            channel_colors=pick_channel_colors(sele_name, channels),
            outline=outline,
    )

pymol.cmd.extend('voxelize', voxelize)
cmd.auto_arg[0]['voxelize'] = cmd.auto_arg[0]['zoom']

def render_view(
        *,
        obj_names,
        atoms_i,
        img_params,
        channel_colors,
        axes=False,
        outline=False,
        frame_ix=None,
        state=-1,
):
    if frame_ix is not None:
        atoms_x = mmdf.transform_atom_coords(atoms_i, frame_ix)
        frame_xi = mmdf.invert_coord_frame(frame_ix)
    else:
        atoms_x = atoms_i
        frame_xi = None

    img = image_from_atoms(atoms_x, img_params)
    render_image(
            obj_names=obj_names,
            img=img,
            grid=img_params.grid,
            channel_colors=channel_colors,
            axes=axes,
            outline=outline,
            frame_xi=frame_xi,
            state=state,
    )

def render_image(
        *,
        obj_names,
        img,
        grid,
        channel_colors,
        axes=False,
        outline=False,
        frame_xi=None,
        state=-1,
):
    view = cmd.get_view()

    # Important to render the axes before the voxels.  I don't know why, but if 
    # the voxels are rendered first, PyMOL regards them as opaque (regardless 
    # of the `transparency_mode` setting.
    if axes:
        ax = cgo_axes()
        cmd.delete(obj_names['axes'])
        cmd.load_cgo(ax, obj_names['axes'])

    if outline:
        edges = cgo_cube_edges(grid.center_A, grid.length_A, outline)
        cmd.delete(obj_names['outline'])
        cmd.load_cgo(edges, obj_names['outline'])

    if img is not None:
        # If `transparency_mode` is disabled (which is the default), CGOs will 
        # be opaque no matter what.
        cmd.set('transparency_mode', 1)

        voxels = cgo_voxels(img, grid, channel_colors)
        cmd.delete(obj_names['voxels'])
        cmd.load_cgo(voxels, obj_names['voxels'])

    if frame_xi is not None:
        for obj in obj_names.values():
            frame_1d = frame_xi.flatten().tolist()
            cmd.set_object_ttt(obj, frame_1d, state)

    cmd.set_view(view)

def select_view(name, sele, grid, frame_ix=None):
    indices = []
    cmd.iterate(
            selection=sele,
            expression='indices.append(index)',
            space=locals(),
    )

    coords_i = np.zeros((len(indices), 4))
    i_from_index = {x: i for i, x in enumerate(indices)}
    cmd.iterate_state(
            selection=sele,
            expression='coords_i[i_from_index[index]] = (x, y, z, 1)',
            space=locals(),
            state=1,
    )

    if frame_ix is not None:
        coords_x = mmdf.transform_coords(coords_i, frame_ix)
    else:
        coords_x = coords_i

    coords_x = coords_x[:,:3] - grid.center_A
    half_len = grid.length_A / 2
    within_grid = np.logical_and(
            coords_x >= -half_len,
            coords_x <= half_len,
    ).all(axis=1)

    cmd.alter(sele, 'b = within_grid[i_from_index[index]]', space=locals())
    cmd.select(name, 'b = 1')

def parse_channels(channels_str):
    return channels_str.split(',') + ['.*']

def parse_element_radius_A(element_radius_A, resolution_A):
    if element_radius_A is None:
        return resolution_A / 2
    else:
        return float(element_radius_A)

def pick_channel_colors(sele, channels):
    elem_colors = []
    cmd.iterate(
            sele,
            'elem_colors.append(dict(element=elem, color=color))',
            space=locals(),
    )

    elem_colors = pl.DataFrame(elem_colors)
    color_channels = set_atom_channels_by_element(
            elem_colors,
            channels,
            first_match=True,
    )
    most_common_colors = dict(
            color_channels
            .explode('channels')
            .groupby('channels', 'color')
            .len()
            .group_by('channels')
            .agg(pl.all().sort_by('len').last())
            .select('channels', 'color')
            .iter_rows()
    )

    colors = []
    for channel in range(len(channels)):
        try:
            color_i = most_common_colors[channel]
            rgb = cmd.get_color_tuple(color_i)
        except KeyError:
            rgb = (1, 1, 1)

        colors.append(rgb)

    return colors

def cgo_voxels(img, grid, channel_colors=None):
    c, w, h, d = img.shape
    voxels = []

    alpha = get_alpha(img)
    face_masks = pick_faces(alpha)

    if channel_colors is None:
        from matplotlib.cm import tab10
        channel_colors = tab10.colors[:c]
    if len(channel_colors) != c:
        raise ValueError(f"Image has {c} channels, but only {len(channel_colors)} were specified")

    for i, j, k in product(range(w), range(h), range(d)):
        if alpha[i, j, k] == 0:
            continue

        voxels += cgo_cube(
                get_voxel_center_coords(grid, np.array([i, j, k])),
                grid.resolution_A,
                color=mix_colors(channel_colors, img[:, i, j, k]),
                alpha=alpha[i, j, k],
                face_mask=face_masks[:, i, j, k],
        )

    return voxels

def cgo_cube(center, length, color=(1, 1, 1), alpha=1.0, face_mask=6 * (1,)):
    # The starting point for this function came from the PyMOL wiki:
    #
    # https://pymolwiki.org/index.php/Cubes
    #
    # However, this starting point (i) didn't support color or transparency and 
    # (ii) had some bugs relating to surface normals.

    verts = np.array([
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 0],
            [1, 1, 1],
    ])
    verts = length * (verts - 0.5) + np.array(center)

    # The order in which the vertices are specified is important: it determines 
    # which direction the triangle faces.  Specifically, a triangle is facing 
    # the camera when its vertices appear in counter-clockwise order.
    #
    # https://stackoverflow.com/questions/8142388/in-what-order-should-i-send-my-vertices-to-opengl-for-culling#8142461
    #
    # Cube:
    #   2───6    y
    #  ╱│  ╱│    │
    # 3─┼─7 │    │
    # │ 0─┼─4    o───x
    # │╱  │╱    ╱
    # 1───5    z 
    #
    # Faces:
    #   x     -x      y     -y      z     -z
    # 7───6  2───3  2───6  1───5  3───7  6───2
    # │   │  │   │  │   │  │   │  │   │  │   │
    # │   │  │   │  │   │  │   │  │   │  │   │
    # 5───4  0───1  3───7  0───4  1───5  4───0
    #
    # In all of the triangle fans below, I'll start with the lower-left vertex 
    # (e.g. 0 for the -x face) and continue counter-clockwise.

    def face(normal, indices):
        return [
                BEGIN, TRIANGLE_FAN,
                ALPHA, alpha,
                COLOR, *color,
                NORMAL, *normal,
                VERTEX, *verts[indices[0]],
                VERTEX, *verts[indices[1]],
                VERTEX, *verts[indices[2]],
                VERTEX, *verts[indices[3]],
                END,
        ]

    faces = []
    x, y, z = np.eye(3)

    if face_mask[0]: faces += face(+x, [5, 4, 6, 7])
    if face_mask[1]: faces += face(-x, [0, 1, 3, 2])
    if face_mask[2]: faces += face(+y, [3, 7, 6, 2])
    if face_mask[3]: faces += face(-y, [0, 4, 5, 1])
    if face_mask[4]: faces += face(+z, [1, 5, 7, 3])
    if face_mask[5]: faces += face(-z, [4, 0, 2, 6])

    return faces

def cgo_cube_edges(center, length, color=(1, 1, 1)):
    if color and not isinstance(color, tuple):
        color = (1, 1, 0)

    verts = np.array([
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 0],
            [1, 1, 1],
    ])
    verts = length * (verts - 0.5) + np.array(center)

    #   2───6
    #  ╱│  ╱│
    # 3─┼─7 │
    # │ 0─┼─4
    # │╱  │╱
    # 1───5

    edges = [
            (0, 1), (0, 2), (0, 4),
            (1, 3), (1, 5),
            (2, 3), (2, 6),
            (3, 7),
            (4, 5), (4, 6),
            (5, 7),
            (6, 7),
    ]

    cube = [
            BEGIN, LINES,
            COLOR, *color,
    ]

    for i, j in edges:
        cube += [
                VERTEX, *verts[i],
                VERTEX, *verts[j],
        ]

    cube += [
            END,
    ]

    return cube

def cgo_axes():
    w = 0.06        # cylinder width 
    l1 = 0.75       # cylinder length
    l2 = l1 + 0.25  # cylinder + cone length
    d = w * 1.618   # cone base diameter

    origin = np.zeros(3)
    x, y, z = np.eye(3)
    r, g, b = np.eye(3)

    return [
            CYLINDER, *origin, *(l1 * x), w, *r, *r,
            CYLINDER, *origin, *(l1 * y), w, *g, *g,
            CYLINDER, *origin, *(l1 * z), w, *b, *b,
            CONE, *(l1 * x), *(l2 * x), d, 0, *r, *r, 1, 1,
            CONE, *(l1 * y), *(l2 * y), d, 0, *g, *g, 1, 1,
            CONE, *(l1 * z), *(l2 * z), d, 0, *b, *b, 1, 1,
    ]

def get_alpha(img):
    img = np.sum(img, axis=0)
    return np.clip(img, 0, 1)

def pick_faces(img):
    face_masks = np.ones((6, *img.shape), dtype=bool)

    face_masks[0, :-1] = img[:-1] > img[1:]
    face_masks[1, 1:] = img[1:] > img[:-1]
    face_masks[2, :, :-1] = img[:, :-1] > img[:, 1:]
    face_masks[3, :, 1:] = img[:, 1:] > img[:, :-1]
    face_masks[4, :, :, :-1] = img[:, :, :-1] > img[:, :, 1:]
    face_masks[5, :, :, 1:] = img[:, :, 1:] > img[:, :, :-1]

    return face_masks

def mix_colors(colors, weights=None):
    if weights is None:
        weights = np.ones(len(colors))

    weights = np.array(weights).reshape(-1, 1)
    ratios = weights / np.sum(weights)

    latent_in = np.array([
            mixbox.float_rgb_to_latent(x)
            for x in colors
    ])
    latent_out = np.sum(latent_in * ratios, axis=0)

    return mixbox.latent_to_float_rgb(latent_out)




from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension

setup(
        ext_modules=[
            Pybind11Extension(
                name='macromol_voxelize._voxelize',
                sources=[
                    'macromol_voxelize/_voxelize.cc',
                ],
                include_dirs=[
                        'macromol_voxelize/vendored/Eigen',
                        'macromol_voxelize/vendored/overlap',
                ],
                cxx_std=14,
            ),
        ],
)
